;;; CL-Wav-Synth -  Manipulate WAV files
;;;
;;; Copyright (C) 2006 Philippe Brochard (hocwp@free.fr)
;;;
;;; #Date#: Mon May 29 15:10:08 2006
;;;
;;; **********************************************
;;; The authors grant you the rights to distribute
;;; and use this software as governed by the terms
;;; of the Lisp Lesser GNU Public License
;;; (http://opensource.franz.com/preamble.html),
;;; known as the LLGPL.
;;; **********************************************



(in-package :cl-user)

(defpackage :cl-wav-synth-clim
  (:nicknames :wav-clim)
  (:use :clim :clim-lisp :cl-wav-synth :uni-shell)
  (:export :run :wav-clim :wav-clim-song))

(in-package :clim-listener)

;;; Grab from Clim Listener - Add: Ignore errors in eval
(define-command (com-eval :menu t :command-table lisp-commands)
    ((form 'clim:form :prompt "Form"))
  (let ((values (multiple-value-list
		 (handler-case (eval form)
		   (error (condition) (format nil "~A" condition))))))
    (fresh-line)
    (shuffle-specials form values)
    (display-evalues values)
    (fresh-line)))

;; Test: editing form... (not sure it's the right way, but it works :)
(define-command (com-edit-form :name t :menu t :command-table lisp-commands)
    ((form 'form :prompt "Form"))
  (let ((new-form (accept 'clim:form :stream (frame-standard-input *application-frame*)
			  :prompt "New form" :default form
			  :insert-default t)))
    (present (list 'com-eval new-form) 'command)
    (clim-listener::com-eval new-form)))


(define-presentation-to-command-translator edit-form
    (clim:command com-edit-form lisp-commands
		  :gesture :describe :echo t
		  :documentation ((object stream)
				  (format stream "Edit ~A" (second object)))
		  :pointer-documentation ((object stream)
					  (format stream "Edit ~A" (second object)))
		  :tester ((object) (eql (car object) 'com-eval)))
    (object)
  (list (second object)))



(define-mime-type (audio cl-wav)
    (:extensions "wav"))

(defmethod mime-type-to-command ((mime-type audio/cl-wav) pathname)
  (values `(wav-clim::com-load-as-sample ,pathname)
          (format nil "Load as sample ~A" (file-namestring pathname))
          (format nil "Load as sample ~A" pathname)))



(define-mime-type (text spectrum)
    (:extensions "spect"))

(defmethod mime-type-to-command ((mime-type text/spectrum) pathname)
  (values `(wav-clim::com-load-as-spectrum ,pathname 100)
          (format nil "Load as spectrum ~A" (file-namestring pathname))
          (format nil "Load as spectrum ~A" pathname)))


(define-mime-type (text song)
    (:extensions "song"))

(defmethod mime-type-to-command ((mime-type text/song) pathname)
  (values `(wav-clim::com-read-song-filename ,pathname)
          (format nil "Load as song ~A" (file-namestring pathname))
          (format nil "Load as song ~A" pathname)))


(in-package :cl-wav-synth-clim)


(defparameter *spi* (float pi 1f0))



(defun bound-value (val min max)
  (max (min val max) min))

(defun build-sample (&optional time)
  (with-application-frame (frame)
    (build-from-freq (base-freqs frame)
		     :n-samples-per-sec (n-samples-per-sec frame)
		     :n-bits-per-sample (n-bits-per-sample frame)
		     :time (or time (build-time frame)))))


(defun set-bound-value (frame)
  (when (sample frame)
    (with-slots (sample-ampl sample-length sample-pos sample-index)
	(find-pane-named frame 'map-sample)
      (let ((maxampl (max (* 2 (loop for i across (data (sample frame))
				     maximize (abs i)))
			  10)))
	(setf sample-index 0)
	(setf sample-pos 0.5)
	(setf sample-length (- (length (data (sample frame))) 2))
	(setf sample-ampl (* maxampl 0.75))))))



(defclass sample-pane (application-pane)
  ((sample-ampl :initarg :sample-ampl :initform 1000 :accessor sample-ampl)
   (sample-length :initarg :sample-length :initform 1000 :accessor sample-length)
   (sample-pos :initarg :sample-pos :initform 0.5 :accessor sample-pos)
   (sample-index :initarg :sample-index :initform 0 :accessor sample-index)))

(defclass freq-pane (application-pane)
  ())

(defclass phase-pane (application-pane)
  ())

(defclass song-pane (application-pane)
  ((song-origine-x :initarg :song-origine-x :initform -1.5 :accessor song-origine-x)
   (song-origine-y :initarg :song-origine-y :initform -1 :accessor song-origine-y)
   (song-scaling :initarg :song-scaling :initform 100 :accessor song-scaling)
   (song-sample-radius :initarg :song-sample-radius :initform 0.1 :accessor song-sample-radius)
   (song-sample-filename :initarg :song-sample-filename :initform "test-song"
			 :accessor song-sample-filename)))

(defclass graphical-view (view)
  ())

(defparameter +graphical-view+ (make-instance 'graphical-view))



(define-application-frame wav-clim (clim-listener::listener)
  ((sample :initarg :sample :initform nil :accessor sample)
   (application :initarg :application :initform 'sample :accessor application)
   (base-freqs :initarg :base-freqs
	       :initform '((100 1 0) (200 0 0) (300 0 0) (400 0 0) (500 0 0))
	       :accessor base-freqs)
   (song :initarg :song :initform nil :accessor song)
   (draw-vertbar :initarg :draw-vertbar :initform t :accessor draw-vertbar)
   (build-time :initarg :build-time :initform (* 3 (/ 100)) :accessor build-time)
   (play-time :initarg :play-time :initform 1 :accessor play-time)
   (n-samples-per-sec :initarg :n-samples-per-sec :initform 8000
		      :accessor n-samples-per-sec)
   (n-bits-per-sample :initarg :n-bits-per-sample :initform 8
		      :accessor n-bits-per-sample))
  (:command-table (wav-clim
		   :inherit-from (clim-listener::application-commands
				  clim-listener::lisp-commands
				  clim-listener::filesystem-commands
				  clim-listener::show-commands
				  file-command-table
				  help-command-table)
		   :menu (("File"          :menu file-command-table)
			  ("Layout"        :menu application-command-table)
			  ("Sample"        :menu build-sample-menu-command-table)
			  ("Song"          :menu build-song-menu-command-table)
			  ("Listener"      :menu listener-command-table)
			  ("Help"          :menu help-command-table))))
  (:panes
   (map-sample (make-pane 'sample-pane :min-heigth 200
			  :scroll-bars nil
			  :foreground +white+ :background +grey+
			  :display-time :command-loop
			  :display-function #'draw-sample))
   (map-freq (make-pane 'freq-pane :min-heigth 100
			:scroll-bars nil
			:foreground +white+ :background +grey+
			:display-time :command-loop
			:display-function #'draw-freqs))
   (map-phase (make-pane 'phase-pane :min-heigth 100
			 :scroll-bars nil
			 :foreground +white+ :background +grey+
			 :display-time :command-loop
			 :display-function #'draw-phase))
   (map-song (make-pane 'song-pane :min-heigth 100 :scroll-bars nil
			:foreground +white+ :background +grey+
			:display-function #'draw-song
			:default-view +graphical-view+))
   (pointer-doc :pointer-documentation)
   (interactor :interactor :min-heigth 200
	       :display-function
	       #'clim-listener::listener-initial-display-function))
  (:layouts
   (default
       (vertically ()
	 (1/2 map-sample)
	 (1/2 interactor)
	 pointer-doc))
   (view-sample
    (vertically ()
      (1/2 map-sample)
      (1/2 interactor)
      pointer-doc))
   (view-sample-1-3
    (vertically ()
      (1/3 map-sample)
      (2/3 interactor)
      pointer-doc))
   (view-sample-ni
    (vertically ()
      map-sample
      pointer-doc))
   (build-sample
    (vertically ()
      (1/3 map-sample)
      (1/6 map-freq)
      (1/6 map-phase)
      (1/3 interactor)
      pointer-doc))
   (build-sample-left
    (vertically ()
      (horizontally ()
	(1/2 (vertically ()
	       (1/2 map-sample)
	       (1/4 map-freq)
	       (1/4 map-phase)))
	(1/2 interactor))
      pointer-doc))
   (build-sample-ni
    (1/2 map-sample)
    (1/4 map-freq)
    (1/4 map-phase))
   (build-song
    (vertically ()
      (1/2 map-song)
      (1/2 interactor)
      pointer-doc))
   (view-listener
    (vertically ()
      interactor
      pointer-doc)))
  (:top-level (default-frame-top-level :prompt 'clim-listener::print-listener-prompt)))





(defmethod frame-standard-output ((frame wav-clim))
  (get-frame-pane frame 'interactor))


(defmethod adopt-frame :after (frame-manager (frame wav-clim))
  (declare (ignore frame-manager))
  (set-bound-value frame))

(defun update-all-panes ()
  (redisplay-frame-panes *application-frame*))



;;; CLim listener internal hack (not sure it's the right way, but it works :)
(make-command-table 'listener-command-table
		    :errorp nil
		    :menu '(("Application" :menu clim-listener::application-commands)
			    ("Lisp"        :menu clim-listener::lisp-commands)
			    ("Filesystem"  :menu clim-listener::filesystem-commands)
			    ("Show"        :menu clim-listener::show-commands)))

(defun add-command-table-to-listener (command-table)
  (map-over-command-table-names (lambda (name command)
				  (when (eql *package* (symbol-package command))
				    (add-command-to-command-table command
								  'clim-listener::application-commands
								  :name name :menu nil :errorp nil)))
				command-table))
;;; ---



;;; Gesture definitions
(define-gesture-name :center :pointer-button (:left))
(define-gesture-name :zoom-in :pointer-button (:left :meta))
(define-gesture-name :zoom-out :pointer-button (:right :meta))

(define-gesture-name :add-sample :pointer-button (:middle))

(define-gesture-name :edit-sample :pointer-button (:middle :control))
(define-gesture-name :move-sample :pointer-button (:left))
(define-gesture-name :copy-sample :pointer-button (:left :control))
(define-gesture-name :play-sample :pointer-button (:right :control))

(define-gesture-name :delete-sample :pointer-button (:middle :meta))




;;; Startup methods
(defun run (&optional (app :sample))
  (let ((application 'sample)
	(layout 'default)
	(draw-vertbar t)
	(sample nil))
    (ecase app
      (:sample nil)
      (:freq (setf application 'build-sample
		   layout 'build-sample
		   draw-vertbar nil
		   sample (build-from-freq '((100 1) (300 0) (500 0))
					   :n-samples-per-sec 8000
					   :n-bits-per-sample 8
					   :time (* 3 (/ 440)))))
      (:song (setf application 'build-song
		   layout 'build-song)))
    (clim-listener::run-frame-top-level
     (make-application-frame 'wav-clim
			     :application application
			     :sample sample
			     :width 800 :height 600
			     :current-layout layout
			     :draw-vertbar draw-vertbar
			     :system-command-reader nil)
     :listener-funcall nil)))




(defgeneric wav-clim (object))

(defmethod wav-clim ((sample sample))
  (clim-listener::run-frame-top-level
   (make-application-frame 'wav-clim
			   :application 'sample
			   :sample sample
			   :width 800 :height 600
			   :current-layout 'view-sample
			   ;;:current-layout 'default
			   :system-command-reader nil)
   :listener-funcall nil))

(defmethod wav-clim ((spectre spectre))
  (clim-listener::run-frame-top-level
   (make-application-frame 'wav-clim
			   :application 'spectre
			   :sample spectre
			   :width 800 :height 600
			   :current-layout 'view-sample
			   :system-command-reader nil)))

(defmethod wav-clim ((base-freqs list))
  (clim-listener::run-frame-top-level
   (make-application-frame 'wav-clim
			   :application 'build-sample
			   :width 800 :height 600
			   :sample (build-from-freq base-freqs
						    :n-samples-per-sec 8000
						    :n-bits-per-sample 8
						    :time (* 3 (/ 440)))
			   :base-freqs base-freqs
			   :current-layout 'build-sample
			   :draw-vertbar nil
			   :system-command-reader nil)))

(defun wav-clim-song (song)
  (clim-listener::run-frame-top-level
   (make-application-frame 'wav-clim
			   :song (copy-list song)
			   :application 'build-song
			   :width 800 :height 600
			   :current-layout 'build-song
			   :system-command-reader nil)))
;;; --- Startup methods
  










;;; Repaint methods
(defun bsubseq (seq &optional (start 0) (end nil))
  "Like as subseq but in [0-length seq]"
  (let* ((s (min (max 0 start)
		 (length seq)))
	 (e (max (min (length seq)
		      (or end (length seq)))
		 s)))
    (subseq seq s e)))


(defun draw-sample (frame pane)
  (labels ((display-time (x y val align-x align-y)
	     (draw-text* pane (format nil "~,2Fs (~D)"
				      (sample-to-time (sample-n-samples-per-sec (sample frame)) val)
				      (truncate val))
			 x y :align-x align-x :align-y align-y :ink +yellow+)
	     (force-output pane))
	   (display-ampl (x y val align-x align-y)
	     (draw-text* pane (format nil "~,2F (~D)"
				      (/ val (sample-max-ampl (sample frame)))
				      (truncate val))
			 x y :align-x align-x :align-y align-y :ink +yellow+)
	     (force-output pane)))
    (unless (sample frame)
      (setf (sample frame) (null-sample 0.001)))
    (let* ((width (bounding-rectangle-width pane))
	   (height (bounding-rectangle-height pane))
	   (zero-y (* height (sample-pos pane)))
	   (length (truncate (sample-length pane)))
	   (index (truncate (sample-index pane)))
	   (ampl (sample-ampl pane))
	   (array (bsubseq (data (sample frame)) index
			   (+ index length)))
	   (len (length array))
	   (dx (/ width (1- len)))
	   (idx (truncate (/ dx)))
	   (dy (/ height ampl 2)))
      (draw-line* pane 0 zero-y width zero-y :ink +black+)
      (if (<= idx 1)
	  (loop for i below (1- len) do
		(draw-line* pane
			    (* i dx) (- zero-y (* dy (aref array i)))
			    (* (1+ i) dx) (- zero-y (* dy (aref array (1+ i))))
			    :ink +blue+)
		(when (draw-vertbar *application-frame*)
		  (draw-line* pane
			      (* i dx) zero-y
			      (* i dx) (- zero-y (* dy (aref array i)))
			      :ink +red+)))
	  (loop for i below width do
		(loop for j below idx
		      maximize (aref array (+ (* i idx) j)) into max
		      minimize (aref array (+ (* i idx) j)) into min
		      finally (draw-line* pane
					  i (- zero-y (* dy max))
					  i (- zero-y (* dy min))
					  :ink +blue+))))
      (display-time 0 height index :left :bottom)
      (display-time width height (+ index (sample-length pane)) :right :bottom)
      (display-ampl 0 0 ampl :left :top))))



(defun draw-freqs (frame pane)
  (let* ((width (bounding-rectangle-width pane))
	 (height (bounding-rectangle-height pane))
	 (len (length (base-freqs frame)))
	 (dx (/ width (1+ len))))
    (draw-line* pane 0 0 width 0 :ink +black+)
    (draw-line* pane 0 (1- height) width (1- height) :ink +black+)
    (loop for freq in (base-freqs frame)
	  for i from 1 do
	  (draw-line* pane
		      (* i dx)  height
		      (* i dx)  (* height (- 1 (second freq))) :ink +red+)
	  (draw-text* pane (format nil "~,3F (~,1FdB)" (second freq)
				   (if (zerop (second freq))
				       -1
				       (* 20 (log (second freq)))))
		      (* i dx)  (* height (- 1 (second freq)))
		      :ink +white+ :align-x :center :align-y :top))))

(defun draw-phase (frame pane)
  (let* ((width (bounding-rectangle-width pane))
	 (height (bounding-rectangle-height pane))
	 (height/2 (/ height 2))
	 (len (length (base-freqs frame)))
	 (dx (/ width (1+ len))))
    (draw-line* pane 0 height/2 width height/2 :ink +black+)
    (loop for freq in (base-freqs frame)
	  for i from 1
	  do
	  (draw-line* pane
		      (* i dx)  height/2
		      (* i dx) (- height/2 (* height/2 (/ (third freq) *spi*)))
		      :ink +green+)
	  (draw-text* pane (format nil "~,1F" (* 180 (/ (third freq) *spi*)))
		      (* i dx) (- height/2 (* height/2 (/ (third freq) *spi*)))
		      :ink +white+ :align-x :center :align-y :top))))

;;; --- Repaint methods



;;; Handle Motion methods
(defun handle-point-freq-pane (pane x y)
  (let* ((frame (pane-frame pane))
	 (width (bounding-rectangle-width pane))
	 (height (bounding-rectangle-height pane))
	 (len (length (base-freqs frame)))
	 (dx (/ width (1+ len)))
	 (index (bound-value (truncate (- (/ x dx) 0.5)) 0 (1- len))))
    (setf (second (nth index (base-freqs frame)))
	  (float (- 1 (/ y height))))
    (setf (sample frame) (build-sample))
    (update-all-panes)))


(defun handle-point-phase-pane (pane x y)
  (let* ((frame (pane-frame pane))
	 (width (bounding-rectangle-width pane))
	 (height (bounding-rectangle-height pane))
	 (height/2 (/ height 2))
	 (len (length (base-freqs frame)))
	 (dx (/ width (1+ len)))
	 (index (bound-value (truncate (- (/ x dx) 0.5)) 0 (1- len))))
    (setf (third (nth index (base-freqs frame)))
	  (float (* (- 1 (/ y height/2)) *spi*)))
    (setf (sample frame) (build-sample))
    (update-all-panes)))



(defun handle-pane-motion (pane x1 y1 handle-point-fun)
  (funcall handle-point-fun pane x1 y1)
  (block processor
    (tracking-pointer
	(pane)
      (:pointer-motion (&key window x y)
		       (declare (ignore window))
		       (funcall handle-point-fun pane x y))
      (:pointer-button-release (&key event x y)
			       (when (= (pointer-event-button event)
					+pointer-left-button+)
				 (funcall handle-point-fun pane x y)
				 (return-from processor nil))))))


(define-presentation-to-command-translator freq-motion
    (blank-area com-freq-motion wav-clim
                :gesture :select :echo t
                :tester ((window) (typep window 'freq-pane)))
    (x y window)
  (list x y window))

(define-wav-clim-command (com-freq-motion :name t :menu nil)
    ((x 'real) (y 'real) (window 'freq-pane))
  (handle-pane-motion window x y #'handle-point-freq-pane))



(define-presentation-to-command-translator phase-motion
    (blank-area com-phase-motion wav-clim
                :gesture :select :echo t
                :tester ((window) (typep window 'phase-pane)))
    (x y window)
  (list x y window))

(define-wav-clim-command (com-phase-motion :name t :menu nil)
    ((x 'real) (y 'real) (window 'phase-pane))
  (handle-pane-motion window x y #'handle-point-phase-pane))



(define-presentation-to-command-translator sample-reset-bound-values
    (blank-area com-sample-reset-bound-values wav-clim
                :gesture nil
		:echo t
                :tester ((window) (typep window 'sample-pane)))
    (window)
  (list window))

(define-wav-clim-command (com-sample-reset-bound-values :name t :menu nil)
    ((window 'sample-pane))
  (declare (ignore window))
  (com-reset-bound-values)
  (update-all-panes))



(define-wav-clim-command (com-find-sample-time-amplitude :name t :menu nil)
    ((x 'real) (y 'real) (window 'sample-pane))
  (let* ((width (bounding-rectangle-width window))
	 (height (bounding-rectangle-height window))
	 (ltext "Plop")
	 (index (sample-index window))
	 (len (sample-length window))
	 (ampl (sample-ampl window))
	 (pos (sample-pos window))
	 (max-ampl (sample-max-ampl (sample *application-frame*))))
    (labels ((convert-to-index (x)
	       (truncate (+ (* x (/ len width)) index)))
	     (convert-to-ampl (y)
	       (truncate (* (- (* height pos) y) (/ ampl height 0.5))))
	     (draw-text (x y)
	       (let ((nx (convert-to-index x))
		     (ny (convert-to-ampl y)))
		 (draw-text* window ltext width 0
			     :ink (pane-background window) :align-x :right :align-y :top)
		 (setf ltext (format nil "Time: ~,2F (~A)  Amplitude: ~,2F (~A)"
				     (float (sample-to-time (sample-n-samples-per-sec (sample *application-frame*))
							    nx))
				     nx
				     (/ ny max-ampl)
				     ny))
		 (draw-text* window ltext width 0
			     :ink +white+ :align-x :right :align-y :top)
		 (force-output window))))
      (draw-text x y)
      (block processor
	(tracking-pointer (window)
	  (:pointer-motion (&key x y)
			   (draw-text x y))
	  (:pointer-button-release (&key x y)
				   (return-from processor (values (convert-to-index x)
								  (convert-to-ampl y)))))))))


;;; --- Handle motion methods







;;; Sample Zooming
(define-wav-clim-command (com-sample-center :name t :menu nil)
    ((x 'real) (y 'real) (map-sample 'sample-pane))
  (multiple-value-bind (index ampl)
      (com-find-sample-time-amplitude x y map-sample)
    (declare (ignore ampl))
    (with-slots (sample-index sample-length) map-sample
      (setf sample-index (bound-value (- index (/ sample-length 2))
				       0
				       (- (length (data (sample *application-frame*)))
					  sample-length 2)))))
  (redisplay-frame-panes *application-frame*))

(define-wav-clim-command (com-sample-time-zoom-in :name t :menu nil)
    ((x 'real) (y 'real) (map-sample 'sample-pane))
  (multiple-value-bind (index ampl)
      (com-find-sample-time-amplitude x y map-sample)
    (declare (ignore ampl))
    (with-slots (sample-index sample-length) map-sample
      (setf sample-length (max (/ sample-length 2) 2))
      (setf sample-index (bound-value (- index (/ sample-length 2))
				       0
				       (- (length (data (sample *application-frame*)))
					  sample-length 2)))))
  (redisplay-frame-panes *application-frame*))

(define-wav-clim-command (com-sample-time-zoom-out :name t :menu nil)
    ((x 'real) (y 'real) (map-sample 'sample-pane))
  (multiple-value-bind (index ampl)
      (com-find-sample-time-amplitude x y map-sample)
    (declare (ignore ampl))
    (with-slots (sample-index sample-length) map-sample
      (setf sample-length (min (* sample-length 2)
			       (- (length (data (sample *application-frame*))) 2)))
      (setf sample-index (bound-value (- index (/ sample-length 2))
				       0
				       (- (length (data (sample *application-frame*)))
					  sample-length 2)))))
  (redisplay-frame-panes *application-frame*))


(define-presentation-to-command-translator sample-center
    (blank-area com-sample-center wav-clim
                :gesture :center :echo t
		:documentation "Center"
		:pointer-documentation "Center"
                :tester ((window) (typep window 'sample-pane)))
    (x y window)
  (list x y window))

(define-presentation-to-command-translator sample-time-zoom-in
    (blank-area com-sample-time-zoom-in wav-clim
                :gesture :zoom-in :echo t
		:documentation "Time Zoom In"
		:pointer-documentation "Time Zoom In"
                :tester ((window) (typep window 'sample-pane)))
    (x y window)
  (list x y window))

(define-presentation-to-command-translator sample-time-zoom-out
    (blank-area com-sample-time-zoom-out wav-clim
                :gesture :zoom-out :echo t
		:documentation "Time Zoom Out"
		:pointer-documentation "Time Zoom Out"
                :tester ((window) (typep window 'sample-pane)))
    (x y window)
  (list x y window))




;;; File command table
(define-command-table file-command-table)

(add-menu-item-to-command-table 'file-command-table "Filesystem"
				:menu 'clim-listener::filesystem-commands)

(add-menu-item-to-command-table 'file-command-table "File" :divider nil)

(define-command (com-play :name t :menu t
			  :command-table file-command-table)
    ()
  (with-application-frame (frame)
    (when (equal (application frame) 'build-sample)
      (setf (sample frame) (build-sample (play-time frame))))
    (case (application frame)
      ((sample build-sample) (play (sample frame)))
      (build-song (com-build-song)
		  (com-play-song)))
    (when (equal (application frame) 'build-sample)
      (setf (sample frame) (build-sample)))))


(add-menu-item-to-command-table 'file-command-table "Load as" :divider nil)

(define-command (com-load-as-sample :name t :menu "Sample"
				    :command-table file-command-table)
    ((path 'pathname))
  (with-application-frame (frame)
    (format t "Loading: ~A as sample~%" path)
    (ignore-errors
      (setf (sample frame) (read-sample path)))
    (set-bound-value frame)
    (redisplay-frame-panes frame)))

(define-command (com-load-as-spectrum :name t :menu "Spectrum"
				      :command-table file-command-table)
    ((path 'pathname)
     (freq 'number :prompt "Frequency"))
  (with-application-frame (frame)
    (format t "Loading: ~A as spectrum~%" path)
    (ignore-errors
      (setf (base-freqs frame) (read-freq-from-file path freq)))
    (setf (sample frame) (build-sample))
    (redisplay-frame-panes frame)))

(define-command (com-load-as-song :name t :menu "Song"
				  :command-table file-command-table)
    ((path 'pathname))
  (com-read-song-filename path))

(add-menu-item-to-command-table 'file-command-table "Exec" :divider nil)

(define-command (com-exec-from-string :name t :menu "Exec"
				      :command-table file-command-table)
    ((command 'string :prompt "Exec"))
  (format t "Starting: ~A~%" command)
  (ignore-errors
    (ushell:ush command)))


(add-menu-item-to-command-table 'file-command-table "Quit" :divider nil)

(define-command (com-quit :name t :menu t
			  :command-table file-command-table
			  :keystroke (#\q :meta :control))
    ()
  (frame-exit *application-frame*))



(add-command-table-to-listener 'file-command-table)
;;; ---  File command table







;;; Layout commands

(define-command-table application-command-table)

(add-menu-item-to-command-table 'application-command-table "View sample" :divider nil)

(define-command (com-set-view-sample :name t :menu "Normal"
				     :command-table application-command-table)
    ()
  (with-application-frame (frame)
    (setf (application frame) 'sample)
    (setf (draw-vertbar frame) t)
    (setf (frame-current-layout frame) 'view-sample)))

(define-command (com-set-view-sample-1-3
		 :name t :menu "Normal 1/3 2/3"
		 :command-table application-command-table)
    ()
  (with-application-frame (frame)
    (setf (application frame) 'sample)
    (setf (draw-vertbar frame) t)
    (setf (frame-current-layout frame) 'view-sample-1-3)))

(define-command (com-set-view-sample-no-interactor
		 :name t :menu "No interactor"
		 :command-table application-command-table)
    ()
  (with-application-frame (frame)
    (setf (application frame) 'sample)
    (setf (draw-vertbar frame) t)
    (setf (frame-current-layout frame) 'view-sample-ni)))


(add-menu-item-to-command-table 'application-command-table "Build sample" :divider nil)

(define-command (com-set-build-samble :name t :menu "Normal"
				      :command-table application-command-table)
    ()
  (with-application-frame (frame)
    (setf (application frame) 'build-sample)
    (setf (sample frame) (build-sample))
    (setf (draw-vertbar frame) nil)
    (setf (frame-current-layout frame) 'build-sample)))



(define-command (com-set-build-sample-left
		 :name t :menu "Normal (listener on left)"
		 :command-table application-command-table)
    ()
  (with-application-frame (frame)
    (setf (application frame) 'build-sample)
    (setf (sample frame) (build-sample))
    (setf (draw-vertbar frame) nil)
    (setf (frame-current-layout frame) 'build-sample-left)))


(define-command (com-set-build-sample-no-listener
		 :name t :menu "No listener"
		 :command-table application-command-table)
    ()
  (with-application-frame (frame)
    (setf (application frame) 'build-sample)
    (setf (sample frame) (build-sample))
    (setf (draw-vertbar frame) nil)
    (setf (frame-current-layout frame) 'build-sample-ni)))

(add-menu-item-to-command-table 'application-command-table "Build song" :divider nil)

(define-command (com-set-build-song
		 :name t :menu "Song"
		 :command-table application-command-table)
    ()
  (with-application-frame (frame)
    (setf (application frame) 'build-song)
    (setf (frame-current-layout frame) 'build-song)))


(add-menu-item-to-command-table 'application-command-table "Listener" :divider nil)

(define-command (com-set-view-listener
		 :name t :menu "Listener"
		 :command-table application-command-table)
    ()
  (with-application-frame (frame)
    (setf (frame-current-layout frame) 'view-listener)))


(add-command-table-to-listener 'application-command-table)

;;; --- Layout commands



;;; Display table

(define-command-table display-command-table)

(add-menu-item-to-command-table 'display-command-table "Set" :divider nil)

(define-command (com-set-amplitude :name t :menu "Amplitude"
				   :command-table display-command-table)
    ((var 'number :prompt "Amplitude"))
  (with-application-frame (frame)
    (let ((map-sample (find-pane-named frame 'map-sample)))
      (with-slots (sample-ampl) map-sample
	(setf sample-ampl var)))
    (redisplay-frame-panes frame)))

(define-command (com-set-position :name t :menu "Position"
				  :command-table display-command-table)
    ((var 'number :prompt "Position"))
  (with-application-frame (frame)
    (let ((map-sample (find-pane-named frame 'map-sample)))
      (with-slots (sample-pos) map-sample
	(setf sample-pos (bound-value var 0 1))))
    (redisplay-frame-panes frame)))


(define-command (com-set-index :name t :menu "Index"
			       :command-table display-command-table)
    ((var 'number :prompt "Index"))
  (with-application-frame (frame)
    (let ((map-sample (find-pane-named frame 'map-sample)))
      (with-slots (sample-index) map-sample
	(setf sample-index var)))
    (redisplay-frame-panes frame)))

(define-command (com-set-begining :name t :menu "Begining"
				  :command-table display-command-table)
    ((var 'number :prompt "Begining time"))
  (with-application-frame (frame)
    (let ((map-sample (find-pane-named frame 'map-sample)))
      (with-slots (sample-index) map-sample
	(setf sample-index (time-to-sample (sample-n-samples-per-sec (sample *application-frame*))
					   var))))
    (redisplay-frame-panes frame)))


(define-command (com-set-length :name t :menu "Length"
				:command-table display-command-table)
    ((var 'number :prompt "Length"))
  (with-application-frame (frame)
    (let ((map-sample (find-pane-named frame 'map-sample)))
      (with-slots (sample-length) map-sample
	(setf sample-length var)))
    (redisplay-frame-panes frame)))


(define-command (com-set-duration :name t :menu "Duration"
				  :command-table display-command-table)
    ((var 'number :prompt "Duration"))
  (with-application-frame (frame)
    (let ((map-sample (find-pane-named frame 'map-sample)))
      (with-slots (sample-length) map-sample
	(setf sample-length (time-to-sample (sample-n-samples-per-sec (sample *application-frame*))
					    var))))
    (redisplay-frame-panes frame)))


(add-menu-item-to-command-table 'display-command-table "Reset" :divider nil)

(define-command (com-reset-bound-values :name t :menu "Bound values"
					:command-table display-command-table)
    ()
  (with-application-frame (frame)
    (set-bound-value frame)))

(define-command (com-set-all-to-zero :name t :menu "All to zero"
				     :command-table display-command-table)
    ()
  (with-application-frame (frame)
    (loop for freq in (base-freqs frame) do
	  (setf (second freq) 0
		(third freq) 0))
    (setf (sample frame) (build-sample))
    (redisplay-frame-panes frame)))

(add-menu-item-to-command-table 'display-command-table "Display" :divider nil)


(define-command (com-redisplay :name t :menu "Redisplay"
			       :command-table display-command-table)
    ()
  (with-application-frame (frame)
    (redisplay-frame-panes frame)))


(add-command-table-to-listener 'display-command-table)

;;; --- Display table



;;; Build sample command table

(define-command-table build-sample-command-table)

(define-command (com-set-sample-per-sec :name t :menu "Set sample per second"
					:command-table build-sample-command-table)
    ((sample 'number :prompt "Sample Per Second"))
  (with-application-frame (frame)
    (when (numberp sample)
      (setf (n-samples-per-sec frame) sample)
      (setf (sample frame) (build-sample))
      (set-bound-value frame)
      (redisplay-frame-panes frame))))

(define-command (com-set-bits-per-sample :name t :menu "Set bits per second"
					 :command-table build-sample-command-table)
    ((bits 'number :prompt "Bits Per Second"))
  (with-application-frame (frame)
    (when (numberp bits)
      (setf (n-bits-per-sample frame) bits)
      (setf (sample frame) (build-sample))
      (set-bound-value frame)
      (redisplay-frame-panes frame))))


(define-command (com-set-play-time :name t :menu "Set play time"
				   :command-table build-sample-command-table)
    ((time 'number :prompt "Play Time"))
  (with-application-frame (frame)
    (when (numberp time)
      (setf (play-time frame) time)
      (setf (sample frame) (build-sample))
      (set-bound-value frame)
      (redisplay-frame-panes frame))))


(define-command (com-set-build-time :name t :menu "Set build time"
				    :command-table build-sample-command-table)
    ((period 'number :prompt "Build Time"))
  (with-application-frame (frame)
    (when (numberp period)
      (setf (build-time frame)
	    (* period (/ (loop for freq in (base-freqs frame)
			       minimize (first freq)))))
      (setf (sample frame) (build-sample))
      (set-bound-value frame)
      (redisplay-frame-panes frame))))



(add-command-table-to-listener 'build-sample-command-table)


(make-command-table 'set-sample-per-sec-command-table
		    :errorp nil
		    :menu '(("8000" :command (com-set-sample-per-sec 8000))
			    ("11025" :command (com-set-sample-per-sec 11025))
			    ("16000" :command (com-set-sample-per-sec 16000))
			    ("22050" :command (com-set-sample-per-sec 22050))
			    ("44100" :command (com-set-sample-per-sec 44100))
			    ("48000" :command (com-set-sample-per-sec 48000))
			    ("96000" :command (com-set-sample-per-sec 96000))))


(make-command-table 'set-bits-per-sample-command-table
		    :errorp nil
		    :menu '(("8" :command (com-set-bits-per-sample 8))
			    ("16" :command (com-set-bits-per-sample 16))
			    ("32" :command (com-set-bits-per-sample 32))))

(make-command-table 'set-play-time-command-table
		    :errorp nil
		    :menu '(("0.5 sec" :command (com-set-play-time 0.5))
			    ("1 sec" :command (com-set-play-time 1))
			    ("2 sec" :command (com-set-play-time 2))
			    ("5 sec" :command (com-set-play-time 5))
			    ("10 sec" :command (com-set-play-time 10))))

(make-command-table 'set-build-time-command-table
		    :errorp nil
		    :menu '(("0.5 period" :command (com-set-build-time 0.5))
			    ("1 period" :command (com-set-build-time 1))
			    ("2 period" :command (com-set-build-time 2))
			    ("3 period" :command (com-set-build-time 3))
			    ("4 period" :command (com-set-build-time 4))
			    ("5 period" :command (com-set-build-time 5))
			    ("6 period" :command (com-set-build-time 6))
			    ("7 period" :command (com-set-build-time 7))
			    ("8 period" :command (com-set-build-time 8))
			    ("9 period" :command (com-set-build-time 9))
			    ("10 period" :command (com-set-build-time 10))))

(make-command-table 'build-sample-menu-command-table
		    :errorp nil
		    :menu '(("Display"       :menu display-command-table)
			    ("Set samples per second" :menu set-sample-per-sec-command-table)
			    ("Set bits per sample" :menu set-bits-per-sample-command-table)
			    ("Set play time" :menu set-play-time-command-table)
			    ("Set build time" :menu set-build-time-command-table)))

;;; --- Build sample command table


;;; Help command table

(define-command-table help-command-table)

(define-command (com-short-help :name t :menu "Help"
				:command-table help-command-table)
    ()
  (help))

(define-command (com-long-help :name t :menu "Long help"
			       :command-table help-command-table)
    ()
  (long-help))

(add-menu-item-to-command-table 'help-command-table "Test" :divider nil)

(define-command (com-list-all-test :name t :menu "List all tests"
				   :command-table help-command-table)
    ()
  (list-all-test))

(define-command (com-run-test :name t :menu "Run a test"
			      :command-table help-command-table)
    ((num 'number :prompt "Test number"))
  (let ((name (intern (format nil "TEST~A" num))))
    (when (fboundp name)
      (funcall name))))

(define-command (com-print-test :name t :menu "View code test"
				:command-table help-command-table)
    ((num 'number :prompt "Test number"))
  (print-test num))

(add-command-table-to-listener 'help-command-table)



;;;----------------------------------------------------------------------;;;
;;;                                                                      ;;;
;;;          Song application                                            ;;;
;;;                                                                      ;;;
;;;----------------------------------------------------------------------;;;

(defun convert-point-to-pane (x y window)
  "Convert x, y to time, pos in window coordinates"
  (values (float (+ (/ x (song-scaling window)) (song-origine-x window)))
	  (float (+ (/ y (song-scaling window)) (song-origine-y window)))))

(defun convert-pane-to-point (time pos window)
  "Convert time, pos to x, y in window coordinates"
  (values (float (* (- time (song-origine-x window)) (song-scaling window)))
	  (float (* (- pos (song-origine-y window)) (song-scaling window)))))



(make-command-table 'build-song-menu-command-table
		    :errorp nil
		    :menu '(("Show Sample" :menu build-song-show-sample-command-table)
			    ("Edit sample" :menu build-song-edit-command-table)
			    ("Sample Operation" :menu build-song-operation-command-table)
			    ("Display" :menu build-song-display-command-table)
			    ("File" :menu build-song-file-command-table)))



(define-command-table build-song-command-table)


;;; Show sample Command Table
(define-command-table build-song-show-sample-command-table)

(define-command (com-show-song-sample :name t :menu "Sample"
				      :command-table build-song-show-sample-command-table)
    ((song-sample 'song-sample))
  (eval-song-sample-form song-sample)
  (format t "Time: ~As  Position: ~A  Tags: ~S  Color: ~6,'0X  Lenght: ~As~%"
	  (s-time song-sample)
	  (s-pos song-sample)
	  (s-tags song-sample)
	  (s-color song-sample)
	  (s-length song-sample))
  (format t "  ~S~%" (s-form song-sample))
  (force-output))


(define-command (com-play-sample :name t :menu "Play a sample on song"
				 :command-table build-song-show-sample-command-table)
    ((song-sample 'song-sample))
  (let ((sample (eval-song-sample-form song-sample)))
    (when (sample-p sample)
      (play sample))))


(define-command (com-center-on-song-sample :name t :menu "Center on sample"
					   :command-table build-song-show-sample-command-table)
    ((song-sample 'song-sample))
  (with-application-frame (frame)
    (let ((map-song (find-pane-named frame 'map-song)))
      (setf (song-origine-x map-song) (- (s-time song-sample)
					 (/ (bounding-rectangle-width map-song) (song-scaling map-song) 2)))
      (setf (song-origine-y map-song) (- (s-pos song-sample)
					 (/ (bounding-rectangle-height map-song) (song-scaling map-song) 2))))
    (redisplay-frame-panes frame)))

(add-command-table-to-listener 'build-song-show-sample-command-table)


;;; Edit sample command table

(define-command-table build-song-edit-command-table)

(define-command (com-edit-song-sample-time :name t :menu "Time"
					   :command-table build-song-edit-command-table)
    ((song-sample 'song-sample))
  (with-application-frame (frame)
    (let ((new-time (accept 'number :stream (frame-standard-input frame)
			    :prompt "New time" :default (s-time song-sample)
			    :insert-default t)))
      (setf (s-time song-sample) new-time))
    (redisplay-frame-panes frame)))
  
(define-command (com-edit-song-sample-form :name t :menu "Form"
					   :command-table build-song-edit-command-table)
    ((song-sample 'song-sample))
  (with-application-frame (frame)
    (let ((new-form (accept 'form :stream (frame-standard-input frame)
			    :prompt "New form" :default (s-form song-sample)
			    :insert-default t)))
      (setf (s-form song-sample) new-form))
    (eval-song-sample-form song-sample)
    (redisplay-frame-panes frame)))
  
    
(define-command (com-edit-song-sample-pos :name t :menu "Position"
					  :command-table build-song-edit-command-table)
    ((song-sample 'song-sample))
  (with-application-frame (frame)
    (let ((new-pos (accept 'number :stream (frame-standard-input frame)
			   :prompt "New position" :default (s-pos song-sample)
			   :insert-default t)))
      (setf (s-pos song-sample) new-pos))
    (redisplay-frame-panes frame)))
  
    
(define-command (com-edit-song-sample-tags :name t :menu "Tags"
					   :command-table build-song-edit-command-table)
    ((song-sample 'song-sample))
  (with-application-frame (frame)
    (let ((new-tags (accept 'form :stream (frame-standard-input frame)
			    :prompt "New tags" :default (s-tags song-sample)
			    :insert-default t)))
      (setf (s-tags song-sample) new-tags))
    (redisplay-frame-panes frame)))

(define-command (com-edit-song-sample-color :name t :menu "Color"
					    :command-table build-song-edit-command-table)
    ((song-sample 'song-sample))
  (with-application-frame (frame)
    (let ((new-color (accept 'form :stream (frame-standard-input frame)
			     :prompt "New color" :default (s-color song-sample)
			     :insert-default t)))
      (setf (s-color song-sample) new-color))
    (redisplay-frame-panes frame)))

(define-command (com-edit-song-sample-length :name t :menu "Length"
					     :command-table build-song-edit-command-table)
    ((song-sample 'song-sample))
  (with-application-frame (frame)
    (let ((new-length (accept 'form :stream (frame-standard-input frame)
			      :prompt "New length" :default (s-length song-sample)
			      :insert-default t)))
      (setf (s-length song-sample) new-length))
    (redisplay-frame-panes frame)))


(define-command (com-edit-song-sample :name t :menu "Sample"
				      :command-table build-song-edit-command-table)
    ((song-sample 'song-sample))
  (com-edit-song-sample-time song-sample)
  (com-edit-song-sample-form song-sample)
  (com-edit-song-sample-pos song-sample)
  (com-edit-song-sample-tags song-sample)
  (com-edit-song-sample-color song-sample)
  (com-edit-song-sample-length song-sample))

(add-command-table-to-listener 'build-song-edit-command-table)


;;; Sample Operation command table

(define-command-table build-song-operation-command-table)

(defun drag-song-sample (frame song-sample start-x start-y)
  (let* ((map-song (find-pane-named frame 'map-song))
	 (radius (* (song-sample-radius map-song) (song-scaling map-song)))
	 (length/2 (* (/ (s-length song-sample) 2) (song-scaling map-song)))
	 (lx start-x) (ly start-y)
	 (dx 0) (dy 0)
	 (ltext "")
	 (width (bounding-rectangle-width map-song)))
    (labels ((display-text (time pos)
	       (draw-text* map-song ltext width 0 :align-x :right :align-y :top :ink (pane-background map-song))
	       (setf ltext (format nil "~,2F, ~,2F" time pos))
	       (draw-text* map-song ltext width 0 :align-x :right :align-y :top :ink +white+))
	     (draw-sample (x y)
	       (draw-ellipse* map-song (+ (- x dx) length/2) (- y dy)
			      length/2 0  0 radius
			      :ink +flipping-ink+ :filled nil)))
      (multiple-value-bind (time-x pos-y)
	  (convert-pane-to-point (s-time song-sample) (s-pos song-sample) map-song)
	(setf dx (- start-x time-x)
	      dy (- start-y pos-y)))
      (draw-sample lx ly)
      (display-text (s-time song-sample) (s-pos song-sample))
      (multiple-value-bind (x y)
	  (block processor
	    (tracking-pointer (map-song)
	      (:pointer-motion (&key window x y)
			       (declare (ignore window))
			       (draw-sample lx ly)
			       (draw-sample x y)
			       (multiple-value-bind (time pos)
				   (convert-point-to-pane x y map-song)
				 (display-text time pos))
			       (setf lx x ly y))
	      (:pointer-button-release (&key event x y)
				       (declare (ignore event))
				       (return-from processor (values x y)))))
	(multiple-value-bind (time pos)
	    (convert-point-to-pane (- x dx) (- y dy) map-song)
	  (values time pos))))))

(define-command (com-move-song-sample :name t :menu "Move sample"
				      :command-table build-song-operation-command-table)
    ((song-sample 'song-sample)
     (x 'real) (y 'real))
  (with-application-frame (frame)
    (multiple-value-bind (time pos)
	(drag-song-sample frame song-sample x y)
      (setf (s-time song-sample) time)
      (setf (s-pos song-sample) pos))
    (com-show-song-sample song-sample)
    (redisplay-frame-panes frame)))

    
(define-command (com-copy-song-sample :name t :menu "Copy sample"
				      :command-table build-song-operation-command-table)
    ((song-sample 'song-sample)
     (x 'real) (y 'real))
  (with-application-frame (frame)
    (multiple-value-bind (time pos)
	(drag-song-sample frame song-sample x y)
      (push (make-instance 'song-sample :time time :form (s-form song-sample)
			   :pos pos :tags (s-tags song-sample)
			   :color (s-color song-sample)
			   :length (s-length song-sample))
	    (song frame)))
    (redisplay-frame-panes frame)))



(define-command (com-add-song-sample :name t :menu "Add sample"
				     :command-table build-song-operation-command-table)
    ((time 'number :prompt "Time")
     (pos 'number :prompt "Position"))
  (with-application-frame (frame)
    (let ((new-sample (make-instance 'song-sample :time time :pos pos)))
      (com-edit-song-sample-form new-sample)
      (com-edit-song-sample-tags new-sample)
      (com-edit-song-sample-color new-sample)
      (com-edit-song-sample-length new-sample)
      (eval-song-sample-form new-sample)
      (push new-sample (song frame)))
    (redisplay-frame-panes frame)))

(define-command (com-delete-song-sample :name t :menu "Delete sample"
					:command-table build-song-operation-command-table)
    ((song-sample 'song-sample))
  (with-application-frame (frame)
    (setf (song frame) (remove song-sample (song frame)))
    (redisplay-frame-panes frame)))
    
(add-command-table-to-listener 'build-song-operation-command-table)    



;;; Build Song Display command table

(define-command-table build-song-display-command-table)

(define-command (com-set-song-scaling :name t :menu t
				      :command-table build-song-display-command-table)
    ((scaling 'number :prompt "Scaling"))
  (with-application-frame (frame)
    (let ((map-song (find-pane-named frame 'map-song)))
      (setf (song-scaling map-song) scaling))
    (redisplay-frame-panes frame)))

(define-command (com-set-song-origine :name t :menu t
				      :command-table build-song-display-command-table)
    ((dx 'number :prompt "Time Origine")
     (dy 'number :prompt "Position Origine"))
  (with-application-frame (frame)
    (let ((map-song (find-pane-named frame 'map-song)))
      (setf (song-origine-x map-song) dx)
      (setf (song-origine-y map-song) dy))
    (redisplay-frame-panes frame)))

(define-command (com-set-song-sample-radius :name t :menu t
					    :command-table build-song-display-command-table)
    ((radius 'number :prompt "Sample Radius"))
  (with-application-frame (frame)
    (let ((map-song (find-pane-named frame 'map-song)))
      (setf (song-sample-radius map-song) radius))
    (redisplay-frame-panes frame)))

(add-command-table-to-listener 'build-song-display-command-table)



;;; File command table
(define-command-table build-song-file-command-table)

(defun map-song-filename (pane ext)
  (concatenate 'string (song-sample-filename pane) "." ext))

(define-command (com-set-song-filename :name t :menu "Set song filename"
				       :command-table build-song-file-command-table)
    ((filename 'string :prompt "Filename"))
  (with-application-frame (frame)
    (let ((map-song (find-pane-named frame 'map-song)))
      (setf (song-sample-filename map-song) filename))))


(define-command (com-write-song :name t :menu "Write song"
				:command-table build-song-file-command-table)
    ()
  (with-application-frame (frame)
    (let* ((map-song (find-pane-named frame 'map-song))
	   (filename (map-song-filename map-song "song")))
      (format t "Writing ~A~%" filename)
      (write-song filename (song frame)))))

(define-command (com-read-song :name t :menu "Read song"
			       :command-table build-song-file-command-table)
    ()
  (with-application-frame (frame)
    (let* ((map-song (find-pane-named frame 'map-song))
	   (filename (map-song-filename map-song "song")))
      (format t "Reading ~A~%" filename)
      (read-song filename)
      (setf (song frame) *current-song*)
      (redisplay-frame-panes frame))))

(add-command-table-to-listener 'build-song-file-command-table)



;;; Building command table
(define-command (com-build-song :name t :menu "Build song"
				:command-table build-song-menu-command-table)
    ()
  (with-application-frame (frame)
    (let* ((map-song (find-pane-named frame 'map-song))
	   (filename (map-song-filename map-song "wav")))
      (format t "Building to ~A~%" filename)
      (build-song filename (song frame)))))

(define-command (com-build-song-in-interval :name t :menu "Build song in interval"
					    :command-table build-song-menu-command-table)
    ((begin 'number :prompt "Begining time")
     (end 'number :prompt "End time"))
  (with-application-frame (frame)
    (let* ((map-song (find-pane-named frame 'map-song))
	   (filename (map-song-filename map-song "wav")))
      (format t "Building to ~A~%" filename)
      (build-song-in-interval filename (song frame) begin end))))


(define-command (com-play-song :name t :menu "Play song"
			       :command-table build-song-menu-command-table)
    ()
  (with-application-frame (frame)
    (let* ((map-song (find-pane-named frame 'map-song))
	   (filename (map-song-filename map-song "wav")))
      (format t "Playing ~A~%" filename)
      (play filename))))

(add-command-table-to-listener 'build-song-menu-command-table)


(define-command (com-read-song-filename :name t :menu nil
					:command-table build-song-command-table)
    ((pathname 'pathname))
  (let ((filename (namestring pathname)))
    (com-set-song-filename (subseq filename 0 (position #\. filename :from-end t)))
    (com-read-song)))



;;; Zooming functions

(define-presentation-to-command-translator song-sample-center
    (blank-area com-song-sample-center wav-clim
                :gesture :center :echo t
		:documentation "Center"
		:pointer-documentation "Center"
                :tester ((window) (typep window 'song-pane)))
    (x y window)
  (list x y window))


(define-presentation-to-command-translator song-sample-zoom-in
    (blank-area com-song-sample-zoom-in wav-clim
                :gesture :zoom-in :echo t
		:documentation "Zoom In"
		:pointer-documentation "Zoom In"
                :tester ((window) (typep window 'song-pane)))
    (x y window)
  (list x y window))


(define-presentation-to-command-translator song-sample-zoom-out
    (blank-area com-song-sample-zoom-out wav-clim
                :gesture :zoom-out :echo t
		:documentation "Zoom Out"
		:pointer-documentation "Zoom Out"
                :tester ((window) (typep window 'song-pane)))
    (x y window)
  (list x y window))


(define-presentation-to-command-translator song-sample-add-sample
    (blank-area com-song-sample-add-sample wav-clim
                :gesture :add-sample :echo t
		:documentation "Add Sample"
		:pointer-documentation "Add Sample"
                :tester ((window) (typep window 'song-pane)))
    (x y window)
  (list x y window))


(define-presentation-to-command-translator song-sample-build-song
    (blank-area com-build-song wav-clim
                :gesture nil :echo t
		:documentation ((object stream)
				(declare (ignore object))
				(format stream "Build Song to ~S"
					(map-song-filename (find-pane-named *application-frame* 'map-song) "wav")))
                :tester ((window) (typep window 'song-pane)))
    ()
  nil)

(define-presentation-to-command-translator song-sample-play-song
    (blank-area com-play-song wav-clim
                :gesture nil :echo t
		:documentation ((object stream)
				(declare (ignore object))
				(format stream "Play Song ~S"
					(map-song-filename (find-pane-named *application-frame* 'map-song) "wav")))
                :tester ((window) (typep window 'song-pane)))
    ()
  nil)

(define-presentation-to-command-translator song-sample-write-song
    (blank-area com-write-song wav-clim
                :gesture nil :echo t
		:documentation ((object stream)
				(declare (ignore object))
				(format stream "Write Song to ~S"
					(map-song-filename (find-pane-named *application-frame* 'map-song) "song")))
                :tester ((window) (typep window 'song-pane)))
    ()
  nil)


(define-presentation-to-command-translator song-sample-read-song
    (blank-area com-read-song wav-clim
                :gesture nil :echo t
		:documentation ((object stream)
				(declare (ignore object))
				(format stream "Read Song ~S"
					(map-song-filename (find-pane-named *application-frame* 'map-song) "song")))
                :tester ((window) (typep window 'song-pane)))
    ()
  nil)




(defun drag-info-on-position (x y map-song)
  (let* ((lx 0) (ly 0) (ltext "")
	 (width (bounding-rectangle-width map-song)))
    (labels ((display-text (x y)
	       (multiple-value-bind (time pos)
		   (convert-point-to-pane x y map-song)
		 (draw-text* map-song ltext width 0 :align-x :right :align-y :top :ink (pane-background map-song))
		 (setf ltext (format nil "~,2Fs, ~,2F (scale=~A)" time pos (float (song-scaling map-song))))
		 (draw-text* map-song ltext width 0 :align-x :right :align-y :top :ink +white+)
		 (force-output map-song))))
      (display-text x y)
      (block processor
	(tracking-pointer (map-song)
	  (:pointer-motion (&key window x y)
			   (declare (ignore window))
			   (display-text x y)
			   (setf lx x ly y))
	  (:pointer-button-release (&key event x y)
				   (declare (ignore event))
				   (return-from processor
				     (convert-point-to-pane x y map-song))))))))


(define-wav-clim-command (com-song-sample-center :name t :menu nil)
    ((x 'real) (y 'real) (map-song 'song-pane))
  (multiple-value-bind (zx zy)
      (drag-info-on-position x y map-song)
    (setf (song-origine-x map-song) (- zx (/ (bounding-rectangle-width map-song) (song-scaling map-song) 2)))
    (setf (song-origine-y map-song) (- zy (/ (bounding-rectangle-height map-song) (song-scaling map-song) 2))))
  (redisplay-frame-panes *application-frame*))



(define-wav-clim-command (com-song-sample-zoom-in :name t :menu nil)
    ((x 'real) (y 'real) (window 'song-pane))
  (multiple-value-bind (zx zy)
      (drag-info-on-position x y window)
    (setf (song-scaling window) (* (song-scaling window) 2))
    (setf (song-origine-x window) (- zx (/ (bounding-rectangle-width window) (song-scaling window) 2)))
    (setf (song-origine-y window) (- zy (/ (bounding-rectangle-height window) (song-scaling window) 2))))
  (redisplay-frame-panes *application-frame*))

(define-wav-clim-command (com-song-sample-zoom-out :name t :menu nil)
    ((x 'real) (y 'real) (window 'song-pane))
  (multiple-value-bind (zx zy)
      (drag-info-on-position x y window)
    (setf (song-scaling window) (/ (song-scaling window) 2))
    (setf (song-origine-x window) (- zx (/ (bounding-rectangle-width window) (song-scaling window) 2)))
    (setf (song-origine-y window) (- zy (/ (bounding-rectangle-height window) (song-scaling window) 2))))
  (redisplay-frame-panes *application-frame*))


(define-wav-clim-command (com-song-sample-add-sample :name t :menu nil)
    ((x 'real) (y 'real) (window 'song-pane))
  (multiple-value-bind (zx zy)
      (drag-info-on-position x y window)
    (com-add-song-sample zx zy)))





(add-command-table-to-listener 'build-song-command-table)


;;; accept method for a song-sample presentation (in any view mode):
(define-presentation-method accept ((type song-sample) stream view &key)
  (declare (ignore view))
  (values
   (completing-from-suggestions (Stream :partial-completers '(#\Space))
     (dolist (sample (song *application-frame*))
       (suggest (format nil "~A ~A " (s-time sample) (s-form sample)) sample)))))


;;;;; Presentation in text mode
(define-presentation-method present (song-sample (type song-sample) stream
						 (view textual-view) &key)
  (format stream "~A ~A" (s-time song-sample) (s-form song-sample)))

(define-presentation-method present (song-sample (type song-sample) stream
						 (view graphical-view) &key)
  (multiple-value-bind (time-x pos-y)
      (convert-pane-to-point (s-time song-sample) (s-pos song-sample) stream)
    (let ((length/2-x (* (s-length song-sample) 0.5 (song-scaling stream)))
	  (radius-y (* (song-sample-radius stream) (song-scaling stream))))
      (when (and (>= (+ time-x (* length/2-x 2)) 0)
		 (<= time-x (bounding-rectangle-width stream))
		 (>= (+ pos-y radius-y) 0)
		 (<= (- pos-y radius-y) (bounding-rectangle-width stream)))
	(draw-ellipse* stream
		       (+ time-x length/2-x) pos-y
		       length/2-x 0  0 radius-y
		       :ink (make-rgb-color (/ (ldb (byte 8 16) (s-color song-sample)) #xFF)
					    (/ (ldb (byte 8 8) (s-color song-sample)) #xFF)
					    (/ (ldb (byte 8 0) (s-color song-sample)) #xFF)))
	(draw-ellipse* stream
		       (+ time-x length/2-x) pos-y
		       length/2-x 0  0 radius-y
		       :ink +black+ :filled nil)))))



;;; Drawing function
(defun draw-song (frame pane)
  (labels ((display-text (x y val align-x align-y)
	     (draw-text* pane (format nil "~,2F" val)
			 x y :align-x align-x :align-y align-y :ink +black+))
	   (display-text-2 (x y time pos align-x align-y)
	     (draw-text* pane (format nil "~,2F, ~,2F" time pos)
			 x y :align-x align-x :align-y align-y :ink +black+)))
    (let ((width (bounding-rectangle-width pane))
	  (height (bounding-rectangle-height pane))
	  (orig-x (song-origine-x pane))
	  (orig-y (song-origine-y pane))
	  (scaling (song-scaling pane)))
      (display-text-2 0 0 orig-x orig-y :left :top)
      (display-text-2 0 height orig-x (+ orig-y (/ height scaling)) :left :bottom)
      (display-text width height (+ orig-x (/ width scaling)) :right :bottom)
      (display-text 0 (/ height 2) (+ orig-y (/ height scaling 2)) :left :center)
      (display-text (/ width 2) height (+ orig-x (/ width scaling 2)) :center :bottom))
    (dolist (sample (song frame))
      (present sample 'song-sample :stream pane))))



(defmacro define-song-sample-presentation (name gesture documentation &optional pointer-documentation)
  `(define-presentation-to-command-translator ,name
    (song-sample ,(intern (concatenate 'string "COM-" (string name))) wav-clim
     :gesture ,gesture
     :documentation ,documentation
     :pointer-documentation ,pointer-documentation)
    (object)
    (list object)))

(define-song-sample-presentation center-on-song-sample nil
  "Center on this sample.")

(define-song-sample-presentation play-sample :play-sample
  "Play this sample."
  ((object stream)
   (format stream "Play ~A ~A    " (s-time object) (s-form object))))

(define-song-sample-presentation edit-song-sample :edit-sample
  "Edit this sample."
  ((object stream)
   (format stream "Edit ~A ~A    " (s-time object) (s-form object))))


(define-presentation-to-command-translator move-song-sample
    (song-sample com-move-song-sample wav-clim
		 :gesture :move-sample
		 :documentation "Move this sample"
		 :pointer-documentation ((object stream)
					 (format stream "Move ~A ~A   " (s-time object) (s-form object))))
    (object x y)
  (list object x y))

(define-presentation-to-command-translator copy-song-sample
    (song-sample com-copy-song-sample wav-clim
		 :gesture :copy-sample
		 :documentation "Copy this sample"
		 :pointer-documentation ((object stream)
					 (format stream "Copy ~A ~A   " (s-time object) (s-form object))))
    (object x y)
  (list object x y))


(define-song-sample-presentation delete-song-sample :delete-sample
  "Delete this sample."
  ((object stream)
   (format stream "Delete ~A ~A    " (s-time object) (s-form object))))


(define-song-sample-presentation edit-song-sample-time nil
  "Edit sample time")

(define-song-sample-presentation edit-song-sample-form nil
  "Edit sample form")

(define-song-sample-presentation edit-song-sample-pos nil
  "Edit sample position")

(define-song-sample-presentation edit-song-sample-tags nil
  "Edit sample tags")

(define-song-sample-presentation edit-song-sample-color nil
  "Edit sample color")

(define-song-sample-presentation edit-song-sample-length nil
  "Edit sample length")



;;;----------------------------------------------------------------------

;;; Insert functions

(in-package :cl-wav-synth)

(define-wav-function s-add (&key (time 0) form (pos 0) tags (color #x00FF00) (length 0.2))
  "Add a sample on song"
  (let ((new-sample (make-instance 'song-sample :time time
				   :form form :pos pos :tags tags :color color :length length)))
    (eval-song-sample-form new-sample)
    (push new-sample (cl-wav-synth-clim::song clim::*application-frame*))
    new-sample))

(define-wav-function s-copy (song-sample &key time form pos tags color length)
  "Copy a sample in a song"
  (s-add :time (or time (s-time song-sample))
	 :form (or form (s-form song-sample))
	 :pos (or pos (s-pos song-sample))
	 :tags (or tags (s-tags song-sample))
	 :color (or color (s-color song-sample))
	 :length (or length (s-length song-sample))))


(define-wav-function s-delete (&key time form pos tags color length all)
  "Delete a sample in a song"
  (setf (cl-wav-synth-clim::song clim::*application-frame*)
	(remove-if #'(lambda (x)
		       (or all
			   (and (if time (= (s-time x) time) t)
				(if form (equal (s-form x) form) t)
				(if pos (= (s-pos x) pos) t)
				(if tags (equal (s-tags x) tags) t)
				(if color (= (s-color x) color) t)
				(if length (= (s-length x) length) t))))
		   (cl-wav-synth-clim::song clim::*application-frame*)))
  nil)



(define-wav-function s-map-fun (&key time form pos tags color length all fun)
  "Map a function over a sample in a song matching time/form/pos/tags/color/length.
   When 'all' is true, map on all samples in the song"
  (mapc #'(lambda (x)
	    (when (or all
		      (and (if time (= (s-time x) time) t)
			   (if form (equal (s-form x) form) t)
			   (if pos (= (s-pos x) pos) t)
			   (if tags (equal (s-tags x) tags) t)
			   (if color (= (s-color x) color) t)
			   (if length (= (s-length x) length) t)))
	      (funcall fun x)))
	(cl-wav-synth-clim::song clim::*application-frame*))
  nil)

(define-wav-macro s-map ((&key time form pos tags color length all) &body body)
  "Apply body on sample matching time/form/pos/tags/color/length.
   Capture intentionally the sample into the variable 'it'
   and sample slots (as time, form, pos and tags) are bounds.
   When 'all' is true, apply on all samples in the song"
  `(s-map-fun :time ,time :form ,form :pos ,pos :tags ,tags :color ,color
    :length ,length :all ,all
    :fun (lambda (it)
	   (with-slots (time form pos tags color length) it
	     ,@body))))


(define-wav-function s-print (song-sample)
  "Print a sample in a song"
  (cl-wav-synth-clim::present song-sample)
  (terpri))

(define-wav-macro with-song (&body body)
  "Apply body on the current loaded song. Capture intentionally the song
  in the variable 'it'. (Note: a song is just a list of song-sample)"
  `(symbol-macrolet ((it (cl-wav-synth-clim::song cl-wav-synth-clim::*application-frame*)))
    ,@body))

(define-wav-macro set-song (&body body)
  "Set the current loaded song with the result of body.
  Capture intentionally the song in the variable 'it'.
  (Note: a song is just a list of song-sample)"
  `(symbol-macrolet ((it (cl-wav-synth-clim::song cl-wav-synth-clim::*application-frame*)))
    (setf it ,@body)))


(define-wav-function build-song-from-match (filename &key time form pos tags color length all)
  "Build a song in filename with samples in a song matching time/form/pos/tags/color/length.
   When 'all' is true, build from all samples in the song"
  (format t "Building in ~S~%" filename)
  (build-song filename
	     (loop for x in (cl-wav-synth-clim::song cl-wav-synth-clim::*application-frame*)
		   when (or all
			    (not (sample-p (eval-song-sample-form x)))
			    (and (if time (= (s-time x) time) t)
				 (if form (equal (s-form x) form) t)
				 (if pos (= (s-pos x) pos) t)
				 (if tags (equal (s-tags x) tags) t)
				 (if color (= (s-color x) color) t)
				 (if length (= (s-length x) length) t)))
		   collect x))
  nil)




(define-wav-macro with-sample (&body body)
  "Apply body on the current loaded sample. Capture intentionally the sample
  in the variable 'it'"
  `(symbol-macrolet ((it (cl-wav-synth-clim::sample cl-wav-synth-clim::*application-frame*)))
    ,@body))


(defparameter *undo-level* 0)

(define-wav-macro set-sample (&body body)
  "Set the current loaded sample with the result of body.
   Capture intentionally the sample in the variable 'it'.
   And save the result with a undo/redo method"
  `(symbol-macrolet ((it (cl-wav-synth-clim::sample cl-wav-synth-clim::*application-frame*)))
    (write-sample (format nil "undo-~A.wav" *undo-level*) it)
    (incf *undo-level*)
    (setf it ,@body)))


(define-wav-function undo (&optional (level nil))
  "Undo the effect of set-sample"
  (write-sample (format nil "undo-~A.wav" *undo-level*)
  		(cl-wav-synth-clim::sample cl-wav-synth-clim::*application-frame*))
  (setf *undo-level* (or level (max (1- *undo-level*) 0)))
  (let ((undo-name (format nil "undo-~A.wav" *undo-level*)))
    (when (probe-file undo-name)
      (setf (cl-wav-synth-clim::sample cl-wav-synth-clim::*application-frame*)
	    (read-sample undo-name))))
  (format t "Level: ~A~%" *undo-level*)
  (force-output)
  *undo-level*)

(define-wav-function redo (&optional (level nil))
  "Redo the effect of set-sample"
  (let ((redo-name (format nil "undo-~A.wav" (or level (1+ *undo-level*)))))
    (when (probe-file redo-name)
      (setf (cl-wav-synth-clim::sample cl-wav-synth-clim::*application-frame*)
	    (read-sample redo-name))
      (setf *undo-level* (or level (1+ *undo-level*)))))
  (format t "Level: ~A~%" *undo-level*)
  (force-output)
  *undo-level*)



(define-wav-function clear-undo (&optional (max-level 100))
  "Clear all undo/redo saved states"
  (dotimes (i max-level)
    (let ((undo-name (format nil "undo-~A.wav" i)))
      (if (probe-file undo-name)
	  (delete-file undo-name)
	  (return))))
  (setf *undo-level* 0)
  (format t "Level: ~A~%" *undo-level*)
  (force-output)
  *undo-level*)


(define-wav-function list-undo (&optional (max-level 100))
  "List all undo/redo saved states"
  (let ((cpt -1))
    (dotimes (i max-level)
      (let ((undo-name (format nil "undo-~A.wav" i)))
	(when (probe-file undo-name)
	  (incf cpt))))
    (format t "Level: ~A / ~A~%" *undo-level* cpt)
    (force-output)
    *undo-level*))
